import unittest
from pathlib import Path
from unittest import skip
from unittest.mock import patch

import dask.array as da
import h5py
import numpy as np
import pytest
import tifffile
from scipy.io import loadmat

from pivtools_cli.config import Config
from pivtools_cli.piv.piv_backend.base import CrossCorrelator
from pivtools_cli.piv.piv_backend.cpu_instantaneous import InstantaneousCorrelatorCPU
from pivtools_cli.piv.piv_backend.factory import make_correlator_backend
from pivtools_cli.tests.helpers import assert_arrays_close


def _fake_inpaint_biharm(self, x: np.ndarray) -> np.ndarray:
    return np.nan_to_num(x, nan=1.0)


class InstantaneousPIVTestCase(unittest.TestCase):

    def setUp(self):
        test_dir = Path(__file__).parent
        config_path = test_dir / "config.yaml"
        self.config = Config(config_path)
        self.config.data["paths"]["base_path"] = str(
            test_dir.parent / "data" / "instantaneous_piv"
        )
        camera_path = self.config.base_path / self.config.cameras[0]
        file_paths = [
            camera_path / (self.config.image_format % 1),
            camera_path / (self.config.image_format.replace("_A", "_B") % 1),
        ]

        self.img_pair = np.stack(
            [
                tifffile.imread(file_paths[0]).astype(self.config.image_dtype),
                tifffile.imread(file_paths[1]).astype(self.config.image_dtype),
            ],
            axis=0,
        )
        C, H, W = self.img_pair.shape
        self.img_pair = self.img_pair.reshape((1, C, H, W))
        self.base_path = Path(self.config.data["paths"]["base_path"]) / "Matlab"
        patcher = patch(
            "pivtools_cli.piv.piv_backend.cpu_instantaneous.InstantaneousCorrelatorCPU._inpaint_nans_biharm",
            new=_fake_inpaint_biharm,
        )

        self._mock_inpaint = patcher.start()

    #    def tearDown(self):
    #        self._mock_inpaint.stop()

    def test_x_peaks(self):
        self.config.data["piv"]["secondary_peak"] = False
        cpu_backend = make_correlator_backend(config=self.config)
        piv_result = cpu_backend.correlate_batch(self.img_pair, self.config)
        for i in range(1, 2):  # len(self.config.window_sizes)):
            peak_x_mat_file = self.base_path / f"MATLAB_peak_loc_x_{i}.mat"

            with h5py.File(peak_x_mat_file, "r") as f:

                matlab_peaks = np.array(f["peak_loc_x"], dtype=np.float32)
                matlab_peaks = np.transpose(matlab_peaks, (2, 1, 0))
                # compare_matrices(
                #    matlab_peaks[0], piv_result.passes[i - 1].peak_loc_x[0]
                # )
                assert_arrays_close(
                    self,
                    matlab_peaks,
                    piv_result.passes[i - 1].peak_loc_x,
                    tol=0.001,
                )

    def test_x_peaks_after_bulk(self):
        self.config.data["piv"]["secondary_peak"] = False
        cpu_backend = make_correlator_backend(config=self.config)
        piv_result = cpu_backend.correlate_batch(self.img_pair, self.config)
        for i in range(1, 3):  # len(self.config.window_sizes) + 1):
            peak_x_mat_file = self.base_path / f"MATLAB_peak_loc_x_after_bulk_{i}.mat"

            with h5py.File(peak_x_mat_file, "r") as f:

                matlab_peaks = np.array(f["peak_loc_x"], dtype=np.float32)
                matlab_peaks = np.transpose(matlab_peaks, (2, 1, 0))
                compare_matrices(
                    matlab_peaks[0], piv_result.passes[i - 1].peak_loc_x_after_bulk[0]
                )
                assert_arrays_close(
                    self,
                    matlab_peaks,
                    piv_result.passes[i - 1].peak_loc_x_after_bulk,
                    tol=0.001,
                )

    def test_y_peaks_after_bulk(self):
        self.config.data["piv"]["secondary_peak"] = False
        cpu_backend = make_correlator_backend(config=self.config)
        piv_result = cpu_backend.correlate_batch(self.img_pair, self.config)
        for i in range(1, 3):  # len(self.config.window_sizes) + 1):
            peak_y_mat_file = self.base_path / f"MATLAB_peak_loc_y  _after_bulk_{i}.mat"

            with h5py.File(peak_y_mat_file, "r") as f:

                matlab_peaks = np.array(f["peak_loc_y"], dtype=np.float32)
                matlab_peaks = np.transpose(matlab_peaks, (2, 1, 0))
                compare_matrices(
                    matlab_peaks[0], piv_result.passes[i - 1].peak_loc_y_after_bulk[0]
                )
                assert_arrays_close(
                    self,
                    matlab_peaks,
                    piv_result.passes[i - 1].peak_loc_y_after_bulk,
                    tol=0.001,
                )

    def test_y_peaks(self):
        self.config.data["piv"]["secondary_peak"] = False
        cpu_backend = make_correlator_backend(config=self.config)
        piv_result = cpu_backend.correlate_batch(self.img_pair, self.config)
        for i in range(1, len(self.config.window_sizes) + 1):
            peak_y_mat_file = self.base_path / f"MATLAB_peak_loc_y_{i}.mat"
            with h5py.File(peak_y_mat_file, "r") as f:

                matlab_peaks = np.array(f["peak_loc_y"], dtype=np.float32)
                matlab_peaks = np.transpose(matlab_peaks, (2, 1, 0))
                assert_arrays_close(
                    self, matlab_peaks, piv_result.passes[i - 1].peak_loc_y, tol=0.0001
                )

    def test_a_prime(self):
        self.config.data["piv"]["secondary_peak"] = False
        cpu_backend = make_correlator_backend(config=self.config)
        piv_result = cpu_backend.correlate_batch(self.img_pair, self.config)
        for i in range(2, len(self.config.window_sizes) + 1):
            peak_y_mat_file = self.base_path / f"MATLAB_A_prime_{i}.mat"
            with h5py.File(peak_y_mat_file, "r") as f:

                matlab_image_a = np.array(f["A_prime"], dtype=np.float32)
                compare_matrices(
                    matlab_image_a.T[0],
                    piv_result.passes[i - 1].image_a_prime[0],
                )
                assert_arrays_close(
                    self,
                    matlab_image_a.T,
                    piv_result.passes[i - 1].image_a_prime,
                    tol=0.01,
                )

    def test_a_prime_subset(self):
        self.config.data["piv"]["secondary_peak"] = False
        cpu_backend = make_correlator_backend(config=self.config)
        piv_result = cpu_backend.correlate_batch(self.img_pair, self.config)
        for i in range(2, 3):  # self.MAX_PASS):  # len(self.config.window_sizes)+1):
            peak_y_mat_file = self.base_path / f"MATLAB_a_prime_subset_{i}.mat"
            with h5py.File(peak_y_mat_file, "r") as f:

                matlab_image_a = np.array(f["A_prime_subset"], dtype=np.float32)
                compare_matrices(
                    matlab_image_a.T[0],
                    piv_result.passes[i - 1].image_a_prime_subset[0],
                )
                assert_arrays_close(
                    self,
                    matlab_image_a.T,
                    piv_result.passes[i - 1].image_a_prime_subset,
                    tol=0.005,
                )

    def test_a(self):
        self.config.data["piv"]["secondary_peak"] = False
        cpu_backend = make_correlator_backend(config=self.config)
        piv_result = cpu_backend.correlate_batch(self.img_pair, self.config)
        for i in range(2, 3):  # self.MAX_PASS):  # len(self.config.window_sizes)+1):
            peak_y_mat_file = self.base_path / f"MATLAB_image_a_{i}.mat"
            with h5py.File(peak_y_mat_file, "r") as f:

                matlab_image_a = np.array(f["A"], dtype=np.float32)
                compare_matrices(
                    matlab_image_a.T[2], piv_result.passes[i - 1].image_a_before[2]
                )
                assert_arrays_close(
                    self,
                    matlab_image_a.T,
                    piv_result.passes[i - 1].image_a_before,
                    tol=0.05,
                )

    def test_a_mesh(self):
        self.config.data["piv"]["secondary_peak"] = False
        cpu_backend = make_correlator_backend(config=self.config)
        piv_result = cpu_backend.correlate_batch(self.img_pair, self.config)
        for i in range(2, 3):  # self.MAX_PASS):  # len(self.config.window_sizes)+1):
            peak_y_mat_file = self.base_path / f"MATLAB_image_mesh_A_{i}.mat"
            with h5py.File(peak_y_mat_file, "r") as f:

                matlab_image_mesh_a = np.array(f["im_mesh_A"], dtype=np.float32)
                matlab_image_mesh_a = np.transpose(matlab_image_mesh_a, (2, 1, 0))
                print(matlab_image_mesh_a.shape)
                print(piv_result.passes[i - 1].im_mesh_A.shape)
                ii, jj = 10, 20
                print(
                    "Python mesh:",
                    piv_result.passes[i - 1].im_mesh_A[ii, jj, 0],
                    piv_result.passes[i - 1].im_mesh_A[ii, jj, 1],
                )
                print("MATLAB mesh:", matlab_image_mesh_a[ii, jj, :])
                compare_matrices(
                    matlab_image_mesh_a.T[2], piv_result.passes[i - 1].im_mesh_A[2] + 1
                )
                assert_arrays_close(
                    self,
                    matlab_image_mesh_a,
                    piv_result.passes[i - 1].im_mesh_A + 1,
                    tol=0.0005,
                )

    def test_peak_height(self):
        self.config.data["piv"]["secondary_peak"] = False
        cpu_backend = make_correlator_backend(config=self.config)
        piv_result = cpu_backend.correlate_batch(self.img_pair, self.config)
        for i in range(1, self.MAX_PASS):  # len(self.config.window_sizes)+1):
            peak_height_mat_file = self.base_path / f"MATLAB_peak_height_{i}.mat"
            with h5py.File(peak_height_mat_file, "r") as f:
                matlab_peaks = np.array(f["peak_height"], dtype=np.float32)

                matlab_peaks = np.transpose(matlab_peaks, (2, 1, 0))
                print("PEAKS", matlab_peaks[0, :, :])
                assert_arrays_close(
                    self, matlab_peaks, piv_result.passes[i - 1].peak_mag, tol=0.0001
                )

    def test_ux_mat(self):
        self.config.data["piv"]["secondary_peak"] = False
        cpu_backend = make_correlator_backend(config=self.config)
        piv_result = cpu_backend.correlate_batch(self.img_pair, self.config)
        for i in range(1, 4):  # len(self.config.window_sizes)+1):
            ux_mat_file = self.base_path / f"MATLAB_ux_mat_{i}.mat"

            with h5py.File(ux_mat_file, "r") as f:
                matlab_ux_mat = np.array(f["ux_mat"], dtype=np.float32)
                matlab_ux_mat = np.transpose(matlab_ux_mat, (1, 0))
                print("UX", matlab_ux_mat[0], piv_result.passes[i - 1].ux_mat[0])
                assert_arrays_close(
                    self, matlab_ux_mat, piv_result.passes[i - 1].ux_mat, tol=0.0001
                )

    def test_uy_mat(self):
        self.config.data["piv"]["secondary_peak"] = False
        cpu_backend = make_correlator_backend(config=self.config)
        piv_result = cpu_backend.correlate_batch(self.img_pair, self.config)
        for i in range(1, 5):  # len(self.config.window_sizes)+1):
            uy_mat_file = self.base_path / f"MATLAB_uy_mat_{i}.mat"
            with h5py.File(uy_mat_file, "r") as f:
                matlab_uy_mat = np.array(f["uy_mat"], dtype=np.float32)
                matlab_uy_mat = np.transpose(matlab_uy_mat, (1, 0))
                compare_matrices(matlab_uy_mat, piv_result.passes[i - 1].uy_mat)
                assert_arrays_close(
                    self, matlab_uy_mat, piv_result.passes[i - 1].uy_mat, tol=1.001
                )

    def test_ux_mat_secondary(self):
        self.config.data["piv"]["secondary_peak"] = True
        cpu_backend = make_correlator_backend(config=self.config)
        piv_result = cpu_backend.correlate_batch(self.img_pair, self.config)
        for i in range(1, 3):  # len(self.config.window_sizes)+1):
            ux_mat_file = self.base_path / f"MATLAB_ux_mat_{i}_secondary.mat"
            with h5py.File(ux_mat_file, "r") as f:
                matlab_ux_mat = np.array(f["ux_mat"], dtype=np.float32)
                matlab_ux_mat = np.transpose(matlab_ux_mat, (1, 0))
                assert_arrays_close(
                    self, matlab_ux_mat, piv_result.passes[i - 1].ux_mat, tol=0.0001
                )

    def test_uy_mat_secondary(self):
        self.config.data["piv"]["secondary_peak"] = True
        cpu_backend = make_correlator_backend(config=self.config)
        piv_result = cpu_backend.correlate_batch(self.img_pair, self.config)
        for i in range(1, 3):  # len(self.config.window_sizes)+1):
            uy_mat_file = self.base_path / f"MATLAB_uy_mat_{i}_secondary.mat"
            with h5py.File(uy_mat_file, "r") as f:
                matlab_uy_mat = np.array(f["uy_mat"], dtype=np.float32)
                matlab_uy_mat = np.transpose(matlab_uy_mat, (1, 0))

                assert_arrays_close(
                    self, matlab_uy_mat, piv_result.passes[i - 1].uy_mat, tol=0.0001
                )
                assert True

    def test_nan(self):
        self.config.data["piv"]["secondary_peak"] = False
        cpu_backend = make_correlator_backend(config=self.config)
        piv_result = cpu_backend.correlate_batch(self.img_pair, self.config)

        for i in range(1, self.MAX_PASS):  # len(self.config.window_sizes)+1):
            nan_file = self.base_path / f"MATLAB_nan_{i}.mat"
            with h5py.File(nan_file, "r") as f:
                matlab_nan = np.array(f["nan_mask"], dtype=bool)

            matlab_nan = np.transpose(matlab_nan, (1, 0))
            python_nan = piv_result.passes[i - 1].nan_mask
            try:
                np.testing.assert_array_equal(matlab_nan, python_nan)
            except AssertionError:
                diff = matlab_nan != python_nan
                idx = np.argwhere(diff)

                print(f"nan_mask mismatch: {len(idx)} elements differ")
                for r, c in idx[:10]:
                    print(
                        f"({r},{c}): MATLAB={matlab_nan[r,c]}, Python={python_nan[r,c]}"
                    )
                raise

    def test_Q_mat_secondary(self):

        self.config.data["piv"]["secondary_peak"] = True
        cpu_backend = make_correlator_backend(config=self.config)
        piv_result = cpu_backend.correlate_batch(self.img_pair, self.config)
        for i in range(1, self.MAX_PASS):  # len(self.config.window_sizes)+1):
            q_mat_file = self.base_path / f"MATLAB_Qmat_{i}.mat"
            with h5py.File(q_mat_file, "r") as f:
                matlab_q_mat = np.array(f["Q_mat"], dtype=np.float32)
                matlab_q_mat = np.transpose(matlab_q_mat, (0, 2, 1))
                assert_arrays_close(
                    self, matlab_q_mat, piv_result.passes[i - 1].Q_mat, tol=0.0001
                )

    def test_delta_ab_gauss(self):

        self.config.data["piv"]["secondary_peak"] = False
        cpu_backend = make_correlator_backend(config=self.config)
        piv_result = cpu_backend.correlate_batch(self.img_pair, self.config)
        for i in range(2, 3):  # self.MAX_PASS):  # len(self.config.window_sizes)+1):
            q_mat_file = self.base_path / f"MATLAB_delta_ab_old_gauss_{i}.mat"
            with h5py.File(q_mat_file, "r") as f:
                matlab_delta_ab_old = np.array(f["delta_ab_old"], dtype=np.float32)
                matlab_delta_ab_old = np.transpose(matlab_delta_ab_old, (2, 1, 0))
                print(matlab_delta_ab_old.shape)

                compare_matrices(
                    matlab_delta_ab_old[0],
                    piv_result.passes[i - 1].delta_ab_old_gauss[0],
                )
                assert_arrays_close(
                    self,
                    matlab_delta_ab_old,
                    piv_result.passes[i - 1].delta_ab_old_gauss,
                    tol=0.0001,
                )

    def test_delta_ab_dense(self):

        self.config.data["piv"]["secondary_peak"] = False
        cpu_backend = make_correlator_backend(config=self.config)
        piv_result = cpu_backend.correlate_batch(self.img_pair, self.config)
        for i in range(2, 3):  # self.MAX_PASS):  # len(self.config.window_sizes)+1):
            q_mat_file = self.base_path / f"MATLAB_delta_ab_dense_{i}.mat"
            with h5py.File(q_mat_file, "r") as f:
                matlab_delta_ab_dense = np.array(f["delta_ab_dense"], dtype=np.float32)
                matlab_delta_ab_dense = np.transpose(matlab_delta_ab_dense, (2, 1, 0))
                for numb in range(1):
                    compare_matrices(
                        matlab_delta_ab_dense[numb],
                        piv_result.passes[i - 1].delta_ab_dense_test[numb],
                    )

                assert_arrays_close(
                    self,
                    matlab_delta_ab_dense,
                    piv_result.passes[i - 1].delta_ab_dense_test,
                    tol=0.001,
                )

    def test_delta_ab_filt(self):

        self.config.data["piv"]["secondary_peak"] = False
        cpu_backend = make_correlator_backend(config=self.config)
        piv_result = cpu_backend.correlate_batch(self.img_pair, self.config)
        for i in range(2, 3):  # self.MAX_PASS):  # len(self.config.window_sizes)+1):
            q_mat_file = self.base_path / f"MATLAB_delta_ab_filt_{i}.mat"
            with h5py.File(q_mat_file, "r") as f:
                matlab_delta_ab_filt = np.array(f["delta_ab_filt"], dtype=np.float32)
                matlab_delta_ab_filt = np.transpose(matlab_delta_ab_filt, (2, 1, 0))
                assert_arrays_close(
                    self,
                    matlab_delta_ab_filt,
                    piv_result.passes[i - 1].delta_ab_filt,
                    tol=0.0001,
                )

    def test_delta_ab_pred_test(self):

        self.config.data["piv"]["secondary_peak"] = False
        cpu_backend = make_correlator_backend(config=self.config)
        piv_result = cpu_backend.correlate_batch(self.img_pair, self.config)
        for i in range(2, 3):  # self.MAX_PASS):  # len(self.config.window_sizes)+1):
            q_mat_file = self.base_path / f"MATLAB_delta_ab_pred_test_{i}.mat"
            with h5py.File(q_mat_file, "r") as f:
                matlab_delta_ab_pred = np.array(f["delta_ab_pred"], dtype=np.float32)
                matlab_delta_ab_pred = np.transpose(matlab_delta_ab_pred, (2, 1, 0))
                print(f"MATLAB shape {i}", matlab_delta_ab_pred.shape)
                # print("MATLAB", matlab_delta_ab_pred.shape)
                # print("PYTHON", piv_result.passes[i - 1].delta_ab_pred_test.shape)
                for numb in range(8):
                    compare_matrices(
                        matlab_delta_ab_pred[numb],
                        piv_result.passes[i - 1].delta_ab_pred_test[numb],
                    )
                assert_arrays_close(
                    self,
                    matlab_delta_ab_pred,
                    piv_result.passes[i - 1].delta_ab_pred_test,
                    tol=0.0001,
                )

    def test_delta_ab_old(self):

        self.config.data["piv"]["secondary_peak"] = False
        cpu_backend = make_correlator_backend(config=self.config)
        piv_result = cpu_backend.correlate_batch(self.img_pair, self.config)
        print("PIV", piv_result.passes[0].delta_ab_old[:, 2, 0])
        for i in range(1, 3):  # self.MAX_PASS):  # len(self.config.window_sizes)+1):
            q_mat_file = self.base_path / f"MATLAB_delta_ab_old_padded_{i}.mat"
            with h5py.File(q_mat_file, "r") as f:
                matlab_delta_ab_old = np.array(f["delta_ab_old"], dtype=np.float32)
                matlab_delta_ab_old = np.transpose(matlab_delta_ab_old, (2, 1, 0))
                print("MATLAB shape", matlab_delta_ab_old.shape)
                # print("MATLAB", matlab_delta_ab_pred.shape)
                # print("PYTHON", piv_result.passes[i - 1].delta_ab_pred_test.shape)
                for numb in range(1):
                    compare_matrices(
                        matlab_delta_ab_old[numb],
                        piv_result.passes[i - 1].delta_ab_old[numb],
                    )
                print(matlab_delta_ab_old[:, 0, 0])
                print(matlab_delta_ab_old[:, 1, 0])
                print(matlab_delta_ab_old[:, 2, 0])

                print(piv_result.passes[i - 1].delta_ab_old[:, 0, 0])
                print(piv_result.passes[i - 1].delta_ab_old[:, 1, 0])
                print(piv_result.passes[i - 1].delta_ab_old[:, 2, 0])

                assert_arrays_close(
                    self,
                    matlab_delta_ab_old,
                    piv_result.passes[i - 1].delta_ab_old,
                    tol=0.01,
                )

    def test_delta_ab_old_not_padded(self):

        self.config.data["piv"]["secondary_peak"] = False
        cpu_backend = make_correlator_backend(config=self.config)
        piv_result = cpu_backend.correlate_batch(self.img_pair, self.config)
        for i in range(1, 3):  # self.MAX_PASS):  # len(self.config.window_sizes)+1):
            q_mat_file = self.base_path / f"MATLAB_delta_ab_old_not_padded_{i}.mat"
            with h5py.File(q_mat_file, "r") as f:
                matlab_delta_ab_old = np.array(f["delta_ab_old"], dtype=np.float32)
                matlab_delta_ab_old = np.transpose(matlab_delta_ab_old, (2, 1, 0))
                print("MATLAB shape", matlab_delta_ab_old.shape)
                # print("MATLAB", matlab_delta_ab_pred.shape)
                # print("PYTHON", piv_result.passes[i - 1].delta_ab_pred_test.shape)
                for numb in range(3):
                    compare_matrices(
                        matlab_delta_ab_old[numb],
                        piv_result.passes[i - 1].delta_ab_old_not_padded[numb],
                    )
                assert_arrays_close(
                    self,
                    matlab_delta_ab_old,
                    piv_result.passes[i - 1].delta_ab_old_not_padded,
                    tol=0.0001,
                )


import numpy as np
import pandas as pd


def compare_matrices(matlab_uy, python_uy, precision=6):
    """
    Compare two UY matrices (MATLAB vs Python) and summarize differences.

    Parameters
    ----------
    matlab_uy : 2D array_like
        MATLAB UY matrix.
    python_uy : 2D array_like
        Python UY matrix (same shape as matlab_uy).
    precision : int
        Number of decimal places to print.

    Returns
    -------
    summary_df : pandas DataFrame
        A table with MATLAB, Python, Diff, and Relative Diff for each element.
    """
    if matlab_uy.shape != python_uy.shape:
        raise ValueError(
            f"Matrices must have the same shape: {matlab_uy.shape} != {python_uy.shape}"
        )

    # Flatten for element-wise comparison
    matlab_flat = matlab_uy.ravel()
    python_flat = python_uy.ravel()

    # Differences
    diff = python_flat - matlab_flat
    rel_diff = np.where(matlab_flat != 0, diff / matlab_flat, 0)

    # Prepare a DataFrame for easy inspection
    summary_df = pd.DataFrame(
        {
            "MATLAB": np.round(matlab_flat, precision),
            "Python": np.round(python_flat, precision),
            "Diff": np.round(diff, precision),
            "RelDiff (%)": np.round(100 * rel_diff, precision),
        }
    )

    # Summary stats
    stats = {
        "Mean Absolute Diff": np.mean(np.abs(diff)),
        "Max Absolute Diff": np.max(np.abs(diff)),
        "Min Absolute Diff": np.min(np.abs(diff)),
        "Mean Relative Diff (%)": np.mean(np.abs(rel_diff)) * 100,
        "Max Relative Diff (%)": np.max(np.abs(rel_diff)) * 100,
        "Min Relative Diff (%)": np.min(np.abs(rel_diff)) * 100,
    }

    print("=== Full Element-wise Comparison ===")
    pd.set_option("display.max_rows", None)

    # Show all columns
    pd.set_option("display.max_columns", None)
    print(summary_df)
    print("\n=== Summary Statistics ===")
    for k, v in stats.items():
        print(f"{k}: {v:.6f}")

    print()
    return summary_df, stats


# Example usage:
# summary, stats = compare_uy_matrices(matlab_uy_mat, piv_result.passes[i-1].uy_mat)



import ctypes
import os

import numpy as np

KERNEL_LANCZOS = 0
KERNEL_GAUSSIAN = 1


class Interp2Custom:
    def __init__(self, lib_path=None):
        if lib_path is None:
            # Use platform-appropriate library extension
            lib_extension = ".dll" if os.name == "nt" else ".so"
            lib_path = os.path.join(
                os.path.dirname(__file__), "../lib", f"libinterp2custom{lib_extension}"
            )
        self.lib = ctypes.CDLL(os.path.abspath(lib_path))

        # Set C function signatures
        self.lib.interp2custom.argtypes = [
            np.ctypeslib.ndpointer(dtype=np.float32, flags="F_CONTIGUOUS"),  # y
            np.ctypeslib.ndpointer(dtype=np.uintp, flags="F_CONTIGUOUS"),  # N
            np.ctypeslib.ndpointer(dtype=np.float32, flags="F_CONTIGUOUS"),  # f_i
            np.ctypeslib.ndpointer(dtype=np.float32, flags="F_CONTIGUOUS"),  # f_j
            np.ctypeslib.ndpointer(
                dtype=np.float32, flags="F_CONTIGUOUS"
            ),  # yi (output)
            ctypes.c_int,  # n_interp
        ]
        self.lib.interp2custom.restype = None

    def interpolate(
        self, y, i, j, kernel_type="lanczos", kernel_size=4, gaussian_std=0.65
    ):
        # -------------------
        # Validate inputs
        # -------------------
        if y.ndim != 2:
            raise ValueError("y must be a 2D array")
        if y.dtype != np.float32:
            y = y.astype(np.float32)
        if i.shape != j.shape:
            raise ValueError("i and j must have the same shape")
        if i.dtype != np.float32:
            i = i.astype(np.float32)
        if j.dtype != np.float32:
            j = j.astype(np.float32)

        # Kernel type as int
        if kernel_type.lower() == "lanczos":
            ktype = KERNEL_LANCZOS
        elif kernel_type.lower() == "gaussian":
            ktype = KERNEL_GAUSSIAN
        else:
            raise ValueError(f"Unknown kernel type '{kernel_type}'")

        self.lib.interp1custom_generatelut(
            ktype, kernel_size, ctypes.c_float(gaussian_std)
        )

        N = np.asfortranarray([y.shape[0], y.shape[1]], dtype=np.uintp)
        n_interp = i.size
        yi = np.asfortranarray(np.zeros_like(i, dtype=np.float32))
        y = np.asfortranarray(y)
        i = np.asfortranarray(i)
        j = np.asfortranarray(j)
        self.lib.interp2custom(y, N, i, j, yi, n_interp)
        return yi



import ctypes
import logging
import os
import sys
import time
import traceback
import warnings
from pathlib import Path
from typing import List, Optional
import cv2
import dask.array as da
import numpy as np
from dask.distributed import get_worker
from scipy.ndimage import gaussian_filter
from scipy.signal import convolve2d

# Try to import line_profiler for detailed profiling
try:
    from line_profiler import profile
except ImportError:
    profile = lambda f: f

# Add src to path for unified imports


from pivtools_core.config import Config
from pivtools_cli.piv.piv_backend.base import CrossCorrelator
from pivtools_cli.piv.piv_result import PIVPassResult, PIVResult
from pivtools_cli.piv.piv_backend.outlier_detection import apply_outlier_detection
from pivtools_cli.piv.piv_backend.infilling import apply_infilling


class InstantaneousCorrelatorCPU(CrossCorrelator):
    @profile
    def __init__(self, config: Config, precomputed_cache: Optional[dict] = None) -> None:
        super().__init__()
        # Use platform-appropriate library extension
        lib_extension = ".dll" if os.name == "nt" else ".so"
        lib_path = os.path.join(
            os.path.dirname(__file__), "../..", "lib", f"libbulkxcorr2d{lib_extension}"
        )
        lib_path = os.path.abspath(lib_path)
        if not os.path.isfile(lib_path):
            raise FileNotFoundError(f"Required library file not found: {lib_path}")
        # Add vcpkg bin directory to DLL search path on Windows
        if os.name == "nt":
            vcpkg_bin = os.path.join(os.environ.get('FFTW_LIB_PATH', '').replace('lib', 'bin'))
            if vcpkg_bin and os.path.isdir(vcpkg_bin):
                os.add_dll_directory(vcpkg_bin)
        self.lib = ctypes.CDLL(lib_path)
        self.lib.bulkxcorr2d.restype = ctypes.c_ubyte
        self.delta_ab_pred = None
        self.delta_ab_old = None
        self.prev_win_size = None
        self.prev_win_spacing = None
        # Updated to use C-contiguous (row-major) arrays
        self.lib.bulkxcorr2d.argtypes = [
            np.ctypeslib.ndpointer(dtype=np.float32, flags="C_CONTIGUOUS"),  # fImageA
            np.ctypeslib.ndpointer(dtype=np.float32, flags="C_CONTIGUOUS"),  # fImageB
            np.ctypeslib.ndpointer(dtype=np.float32, flags="C_CONTIGUOUS"),  # fMask
            np.ctypeslib.ndpointer(dtype=np.int32, flags="C_CONTIGUOUS"),  # nImageSize
            np.ctypeslib.ndpointer(dtype=np.float32, flags="C_CONTIGUOUS"),  # fWinCtrsX
            np.ctypeslib.ndpointer(dtype=np.float32, flags="C_CONTIGUOUS"),  # fWinCtrsY
            np.ctypeslib.ndpointer(dtype=np.int32, flags="C_CONTIGUOUS"),  # nWindows
            np.ctypeslib.ndpointer(
                dtype=np.float32, flags="C_CONTIGUOUS"
            ),  # fWindowWeightA
            ctypes.c_bool,  # bEnsemble
            np.ctypeslib.ndpointer(
                dtype=np.float32, flags="C_CONTIGUOUS"
            ),  # fWindowWeightB
            np.ctypeslib.ndpointer(dtype=np.int32, flags="C_CONTIGUOUS"),  # nWindowSize
            ctypes.c_int,  # nPeaks
            ctypes.c_int,  # iPeakFinder
            np.ctypeslib.ndpointer(
                dtype=np.float32, flags="C_CONTIGUOUS"
            ),  # fPkLocX (output)
            np.ctypeslib.ndpointer(
                dtype=np.float32, flags="C_CONTIGUOUS"
            ),  # fPkLocY (output)
            np.ctypeslib.ndpointer(
                dtype=np.float32, flags="C_CONTIGUOUS"
            ),  # fPkHeight (output)
            np.ctypeslib.ndpointer(
                dtype=np.float32, flags="C_CONTIGUOUS"
            ),  # fSx (output)
            np.ctypeslib.ndpointer(
                dtype=np.float32, flags="C_CONTIGUOUS"
            ),  # fSy (output)
            np.ctypeslib.ndpointer(
                dtype=np.float32, flags="C_CONTIGUOUS"
            ),  # fSxy (output)
            np.ctypeslib.ndpointer(
                dtype=np.float32, flags="C_CONTIGUOUS"
            ),  # fCorrelPlane_Out (output)
        ]
        # Window weights should be C-contiguous with shape (win_height, win_width)
        self.win_weights = [
            np.ascontiguousarray(self._window_weight_fun(win_size, config.window_type))
            for win_size in config.window_sizes
        ]

        # Use precomputed cache if provided, otherwise compute it
        if precomputed_cache is not None:
            self._load_precomputed_cache(precomputed_cache)
        else:
            self._cache_window_padding(config=config)
            self.H, self.W = config.image_shape
            # Cache interpolation grids for performance
            self._cache_interpolation_grids(config=config)

        # Initialize vector masks (will be set in correlate_batch)
        self.vector_masks = []

        # Store pass times for profiling
        self.pass_times = []

    def _load_precomputed_cache(self, cache: dict) -> None:
        """Load precomputed cache data to avoid redundant computation.
        
        :param cache: Dictionary containing precomputed cache data
        :type cache: dict
        """
        # Load window padding cache
        self.win_ctrs_x = cache['win_ctrs_x']
        self.win_ctrs_y = cache['win_ctrs_y']
        self.win_spacing_x = cache['win_spacing_x']
        self.win_spacing_y = cache['win_spacing_y']
        self.win_ctrs_x_all = cache['win_ctrs_x_all']
        self.win_ctrs_y_all = cache['win_ctrs_y_all']
        self.n_pre_all = cache['n_pre_all']
        self.n_post_all = cache['n_post_all']
        self.ksize_filt = cache['ksize_filt']
        self.sd = cache['sd']
        self.G_smooth_predictor = cache['G_smooth_predictor']
        
        # Load image dimensions
        self.H = cache['H']
        self.W = cache['W']
        
        # Load interpolation grids cache
        self.im_mesh = cache['im_mesh']
        self.cached_dense_maps = cache['cached_dense_maps']
        self.cached_predictor_maps = cache['cached_predictor_maps']

    def get_cache_data(self) -> dict:
        """Extract cache data for sharing across workers.
        
        :return: Dictionary containing all cached data
        :rtype: dict
        """
        return {
            'win_ctrs_x': self.win_ctrs_x,
            'win_ctrs_y': self.win_ctrs_y,
            'win_spacing_x': self.win_spacing_x,
            'win_spacing_y': self.win_spacing_y,
            'win_ctrs_x_all': self.win_ctrs_x_all,
            'win_ctrs_y_all': self.win_ctrs_y_all,
            'n_pre_all': self.n_pre_all,
            'n_post_all': self.n_post_all,
            'ksize_filt': self.ksize_filt,
            'sd': self.sd,
            'G_smooth_predictor': self.G_smooth_predictor,
            'H': self.H,
            'W': self.W,
            'im_mesh': self.im_mesh,
            'cached_dense_maps': self.cached_dense_maps,
            'cached_predictor_maps': self.cached_predictor_maps,
        }

    @profile
    def correlate_batch(  # type: ignore[override]
        self, images: np.ndarray, config: Config, vector_masks: List[np.ndarray] | None = None
    ) -> PIVResult:
        """Run PIV correlation on a batch of image pairs with MATLAB-style indexing."""

        N, _, H, W = images.shape

        piv_result_all = PIVResult()
        self.delta_ab_pred = None
        self.delta_ab_old = None

        # Clear pass times for this batch
        self.pass_times = []

        # Use pre-computed vector masks
        self.vector_masks = vector_masks if vector_masks is not None else []

        y_coords = np.arange(self.H, dtype=np.float32)
        x_coords = np.arange(self.W, dtype=np.float32)
        y_mesh, x_mesh = np.meshgrid(y_coords, x_coords, indexing="ij")
        self.im_mesh = np.stack([y_mesh, x_mesh], axis=-1)

        for n in range(N):
            try:
                # Convert images to C-contiguous (row-major) format
                image_a = np.asarray(images[n, 0], dtype=np.float32)
                image_b = np.asarray(images[n, 1], dtype=np.float32)

                if not image_a.flags["C_CONTIGUOUS"]:
                    image_a = np.ascontiguousarray(image_a)
                if not image_b.flags["C_CONTIGUOUS"]:
                    image_b = np.ascontiguousarray(image_b)

                # Pass image_size as [H, W] in C-contiguous format
                image_size = np.ascontiguousarray(np.array([H, W], dtype=np.int32))

                for pass_idx, win_size in enumerate(config.window_sizes):
                    pass_start = time.perf_counter()
                    image_a_prime, image_b_prime, self.delta_ab_pred = (
                        self._predictor_corrector(
                            pass_idx,
                            image_a,
                            image_b,
                            win_type=config.window_type,
                        )
                    )

                    (
                        win_size_arr,
                        n_windows,
                        b_mask,
                        n_peaks,
                        i_peak_finder,
                        b_ensemble,
                        pk_loc_x,
                        pk_loc_y,
                        pk_height,
                        sx,
                        sy,
                        sxy,
                        correl_plane_out,
                    ) = self._set_lib_arguments(
                        config=config,
                        win_size=win_size,
                        pass_idx=pass_idx,
                    )
                    
                    # Ensure images are C-contiguous before passing to C library
                    image_a_prime_c = image_a_prime if image_a_prime.flags["C_CONTIGUOUS"] else np.ascontiguousarray(image_a_prime)
                    image_b_prime_c = image_b_prime if image_b_prime.flags["C_CONTIGUOUS"] else np.ascontiguousarray(image_b_prime)
                    
                    try:
                        error_code = self.lib.bulkxcorr2d(
                            image_a_prime_c,
                            image_b_prime_c,
                            b_mask,
                            image_size,
                            self.win_ctrs_x[pass_idx].astype(np.float32),
                            self.win_ctrs_y[pass_idx].astype(np.float32),
                            n_windows,
                            self.win_weights[pass_idx],
                            b_ensemble,
                            self.win_weights[pass_idx],
                            win_size_arr,
                            int(n_peaks),
                            int(i_peak_finder),
                            pk_loc_x,
                            pk_loc_y,
                            pk_height,
                            sx,
                            sy,
                            sxy,
                            correl_plane_out,
                        )
                    except Exception as e:
                        logging.error(f"    Exception type: {type(e).__name__}")
                        import traceback
                        logging.error(traceback.format_exc())
                        raise

                    if error_code != 0:
                        error_names = {
                            1: "ERROR_NOMEM (out of memory)",
                            2: "ERROR_NOPLAN_FWD (FFT forward plan failed)",
                            4: "ERROR_NOPLAN_BWD (FFT backward plan failed)",
                            8: "ERROR_NOPLAN (general plan error)",
                            9: "ERROR_OUT_OF_BOUNDS (array access out of bounds)"
                        }
                        error_msg = error_names.get(error_code, f"Unknown error code {error_code}")
                        logging.error(f"    bulkxcorr2d returned error code {error_code}: {error_msg}")
                        raise RuntimeError(f"bulkxcorr2d failed with error {error_code}: {error_msg}")
                    
                    n_win_y = int(n_windows[0])
                    n_win_x = int(n_windows[1])
                    mask_bool = b_mask.astype(bool)

                    pk_loc_x[:, mask_bool] = np.nan
                    pk_loc_y[:, mask_bool] = np.nan
                    pk_height[:, mask_bool] = np.nan

                    win_height, win_width = win_size_arr.astype(np.int32)
                    large_disp_mask = (
                        (np.abs(pk_loc_x) > win_width / 4.0)
                        | (np.abs(pk_loc_y) > win_height / 4.0)
                    )
                    pk_loc_x[large_disp_mask] = np.nan
                    pk_loc_y[large_disp_mask] = np.nan
                    pk_height[large_disp_mask] = np.nan

                    # delta_ab_pred[..., 0] = Y-displacement, delta_ab_pred[..., 1] = X-displacement
                    # pk_loc_x is X-displacement, pk_loc_y is Y-displacement
                    pk_loc_x += self.delta_ab_pred[..., 1][np.newaxis, :, :]  # Add X-predictor to X
                    pk_loc_y += self.delta_ab_pred[..., 0][np.newaxis, :, :]  # Add Y-predictor to Y

                    primary_idx = np.zeros((1, n_win_y, n_win_x), dtype=np.intp)
                    ux_mat = np.take_along_axis(pk_loc_x, primary_idx, axis=0)[0]
                    uy_mat = np.take_along_axis(pk_loc_y, primary_idx, axis=0)[0]
                    # Use direct indexing without meshgrid for outlier detection and peak selection
                    n_win_y = int(n_windows[0])
                    n_win_x = int(n_windows[1])
                    peak_choice = np.ones((n_win_y, n_win_x), dtype=np.int32)

                    # Initial peak selection
                    ux_mat = pk_loc_x[0]
                    uy_mat = pk_loc_y[0]

                    nan_mask = np.isnan(ux_mat) | np.isnan(uy_mat)
                    
                    # Apply outlier detection if enabled
                    if config.outlier_detection_enabled:
                        outlier_methods = config.outlier_detection_methods
                        if outlier_methods:
                            # Get primary peak magnitude for peak_mag detection
                            primary_peak_mag_temp = pk_height[0]
                            outlier_mask = apply_outlier_detection(
                                ux_mat, uy_mat, outlier_methods, peak_mag=primary_peak_mag_temp
                            )
                            nan_mask |= outlier_mask

                    if config.secondary_peak:
                        for pk in range(1, n_peaks):
                            # Increment peak_choice for nan_mask locations
                            peak_choice[nan_mask] += 1
                            # Clamp peak_choice to valid range
                            peak_choice = np.clip(peak_choice, 1, n_peaks)
                            # Select new peak for nan_mask locations
                            ux_mat = np.choose(peak_choice - 1, pk_loc_x)
                            uy_mat = np.choose(peak_choice - 1, pk_loc_y)
                            if config.outlier_detection_enabled:
                                outlier_methods = config.outlier_detection_methods
                                if outlier_methods:
                                    primary_peak_mag_temp = np.choose(peak_choice - 1, pk_height)
                                    outlier_mask = apply_outlier_detection(
                                        ux_mat, uy_mat, outlier_methods, peak_mag=primary_peak_mag_temp
                                    )
                                    nan_mask |= outlier_mask
                            if not nan_mask.any():
                                break

                    # Select primary peak magnitude
                    primary_peak_mag = np.choose(peak_choice - 1, pk_height)
                    nan_mask |= np.isnan(primary_peak_mag)

                    nan_mask |= mask_bool
                    nan_mask |= primary_peak_mag < 0.2

                    # Q calculation (peak ratio)
                    shifted_pk_height = np.roll(pk_height, shift=-1, axis=0)
                    shifted_pk_height[-1, :, :] = pk_height[-1, :, :]
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore", category=RuntimeWarning)
                        Q_mat = np.divide(
                            pk_height,
                            shifted_pk_height,
                            out=np.zeros_like(pk_height),
                            where=shifted_pk_height > 0,
                        )

                    Q = np.choose(peak_choice - 1, Q_mat)

                    if nan_mask.any():
                        ux_mat[nan_mask] = np.nan
                        uy_mat[nan_mask] = np.nan
                        primary_peak_mag[nan_mask] = np.nan
                        Q[nan_mask] = 0.0

                    ux_mat[mask_bool] = 0.0
                    uy_mat[mask_bool] = 0.0

                    # Apply infilling for mid-passes or final pass
                    is_final_pass = (pass_idx == len(config.window_sizes) - 1)
                    
                    if is_final_pass:
                        # Final pass infilling (optional)
                        final_infill_cfg = config.infilling_final_pass
                        if final_infill_cfg.get('enabled', True) and np.isnan(ux_mat).any():
                            infill_mask = np.isnan(ux_mat) | np.isnan(uy_mat)
                            ux_mat, uy_mat = apply_infilling(
                                ux_mat, uy_mat, infill_mask, final_infill_cfg
                            )
                    else:
                        # Mid-pass infilling (required for predictor)
                        if np.isnan(ux_mat).any() or np.isnan(uy_mat).any():
                            infill_mask = np.isnan(ux_mat) | np.isnan(uy_mat)
                            mid_infill_cfg = config.infilling_mid_pass
                            ux_mat, uy_mat = apply_infilling(
                                ux_mat, uy_mat, infill_mask, mid_infill_cfg
                            )

                    ux_mat[mask_bool] = 0.0
                    uy_mat[mask_bool] = 0.0
                    peak_choice[nan_mask] = 0

                    ux_mat = np.ascontiguousarray(ux_mat.astype(np.float32))
                    uy_mat = np.ascontiguousarray(uy_mat.astype(np.float32))
                    nan_mask = np.ascontiguousarray(nan_mask)
                    Q = np.ascontiguousarray(Q.astype(np.float32))
                    primary_peak_mag = np.ascontiguousarray(
                        np.where(nan_mask, 0.0, primary_peak_mag.astype(np.float32))
                    )
                    pk_height = np.ascontiguousarray(pk_height.astype(np.float32))

                    # Stack as [Y, X] to match im_mesh structure where [..., 0] = Y and [..., 1] = X
                    # This ensures correct image warping: im_mesh + delta_ab aligns Y with Y and X with X
                    self.delta_ab_old = np.stack([uy_mat, ux_mat], axis=2)
                    pre_y, pre_x = self.n_pre_all[pass_idx]
                    post_y, post_x = self.n_post_all[pass_idx]
                    self.delta_ab_old = np.pad(
                        self.delta_ab_old,
                        ((pre_y, post_y), (pre_x, post_x), (0, 0)),
                        mode="edge",
                    )

                    self.previous_win_spacing = (
                        self.win_spacing_y[pass_idx],
                        self.win_spacing_x[pass_idx],
                    )
                    self.prev_win_size = (n_win_y, n_win_x)

                    pass_result = PIVPassResult(
                        n_windows=np.array([n_win_y, n_win_x], dtype=np.int32),
                        ux_mat=np.copy(ux_mat),
                        uy_mat=np.copy(uy_mat),
                        nan_mask=np.copy(nan_mask),
                        peak_mag=np.copy(pk_height),
                        peak_choice=np.copy(peak_choice),
                        predictor_field=np.copy(self.delta_ab_old),
                        b_mask=b_mask.reshape((n_win_y, n_win_x)).astype(bool),
                        window_size=win_size,
                        win_ctrs_x=self.win_ctrs_x[pass_idx],
                        win_ctrs_y=self.win_ctrs_y[pass_idx],

                    )
                    pass_time = time.perf_counter() - pass_start
                    self.pass_times.append((n, pass_idx, pass_time))
                    piv_result_all.add_pass(pass_result)
                    
                    # Explicit memory cleanup to prevent accumulation
                    # These large intermediate arrays can consume 500+ MB per pass
                    del pk_loc_x, pk_loc_y, pk_height, correl_plane_out
                    del image_a_prime, image_b_prime
                    del sx, sy, sxy, Q_mat
                    # Force garbage collection after last pass to release memory
                    if pass_idx == len(config.window_sizes) - 1:
                        import gc
                        gc.collect()

            except Exception as exc:
                logging.error("Error in correlate_batch for image %d: %s", n, exc)
                logging.error(traceback.format_exc())
                raise

        return piv_result_all

    def _compute_window_centres(
        self, pass_idx: int, config: Config
    ) -> tuple[int, int, np.ndarray, np.ndarray]:
        """
        Compute window centers and spacing for a given pass.
        
        Matches MATLAB logic exactly:
        - win_ctrs_x spans the width dimension (Nx = W = columns)
        - win_ctrs_y spans the height dimension (Ny = H = rows)
        - Window centers are in pixel coordinates (0-based)
        - X corresponds to horizontal (width), Y to vertical (height)

        :param pass_idx: Index of the current pass.
        :type pass_idx: int
        :param config: Configuration object containing window sizes, overlap, and image shape.
        :type config: Config
        :return: Tuple containing window spacing in x and y, and arrays of window center coordinates in x and y.
        :rtype: tuple[int, int, np.ndarray, np.ndarray]
        """
        # Image dimensions: config.image_shape = (H, W) = (rows, cols)
        H, W = config.image_shape
        Ny = H  # Number of rows (height)
        Nx = W  # Number of columns (width)
        
        logging.debug(f"_compute_window_centres pass {pass_idx}:")
        logging.debug(f"  Image shape (H, W) = ({H}, {W})")
        logging.debug(f"  Ny (height/rows) = {Ny}, Nx (width/cols) = {Nx}")
        
        # Window size: config.window_sizes[pass_idx] = (win_height, win_width)
        # This matches MATLAB where wsize(1)=height, wsize(2)=width
        win_height, win_width = config.window_sizes[pass_idx]
        overlap = config.overlap[pass_idx]
        
        logging.debug(f"  Window size (H, W) = ({win_height}, {win_width})")
        logging.debug(f"  Overlap = {overlap}%")

        # Window spacing in pixels
        win_spacing_x = round((1 - overlap / 100) * win_width)
        win_spacing_y = round((1 - overlap / 100) * win_height)
        
        logging.debug(f"  Window spacing (X, Y) = ({win_spacing_x}, {win_spacing_y})")

        # MATLAB: win_ctrs_x = 0.5 + wsize(1)/2 : win_spacing_x : Nx - wsize(1)/2 + 0.5
        # But MATLAB then subtracts 1 before passing to C (1-based to 0-based conversion)
        # So in 0-based indexing: win_ctrs_x = -0.5 + wsize(1)/2 : win_spacing_x : Nx - wsize(1)/2 - 0.5
        # For a 128-pixel window (indices 0-127), center is at 63.5
        # For window starting at pixel 0: center = (0 + 127) / 2 = 63.5
        
        # First window center in X (width dimension) - 0-based array indexing
        first_ctr_x = (win_width - 1) / 2.0  # For 128: (127)/2 = 63.5
        # Last possible window center in X
        last_ctr_x = Nx - (win_width + 1) / 2.0  # For W=4872, win=128: 4872 - 64.5 = 4807.5
        
        # First window center in Y (height dimension) - 0-based array indexing
        first_ctr_y = (win_height - 1) / 2.0
        # Last possible window center in Y
        last_ctr_y = Ny - (win_height + 1) / 2.0
        
        logging.debug(f"  X range: [{first_ctr_x:.2f}, {last_ctr_x:.2f}]")
        logging.debug(f"  Y range: [{first_ctr_y:.2f}, {last_ctr_y:.2f}]")
        
        # Number of windows that fit
        n_win_x = int(np.floor((last_ctr_x - first_ctr_x) / win_spacing_x)) + 1
        n_win_y = int(np.floor((last_ctr_y - first_ctr_y) / win_spacing_y)) + 1
        
        # Ensure at least one window
        n_win_x = max(1, n_win_x)
        n_win_y = max(1, n_win_y)
        
        logging.debug(f"  Number of windows (X, Y) = ({n_win_x}, {n_win_y})")

        # Generate window center arrays using linspace (matches MATLAB's colon operator)
        win_ctrs_x = np.linspace(
            first_ctr_x,
            first_ctr_x + win_spacing_x * (n_win_x - 1),
            n_win_x,
            dtype=np.float32,
        )
        win_ctrs_y = np.linspace(
            first_ctr_y,
            first_ctr_y + win_spacing_y * (n_win_y - 1),
            n_win_y,
            dtype=np.float32,
        )
        
        logging.debug(f"  win_ctrs_x: min={win_ctrs_x.min():.2f}, max={win_ctrs_x.max():.2f}, len={len(win_ctrs_x)}")
        logging.debug(f"  win_ctrs_y: min={win_ctrs_y.min():.2f}, max={win_ctrs_y.max():.2f}, len={len(win_ctrs_y)}")

        return (
            win_spacing_x,
            win_spacing_y,
            np.ascontiguousarray(win_ctrs_x),
            np.ascontiguousarray(win_ctrs_y),
        )

    def _check_args(self, *args):
        """Check the arguments for consistency and validity if debug mode is enabled.
        Parameters
        ----------
        *args : list of tuples
            Each tuple contains (name, array) to be checked.

        """

        def _describe(arr):
            if isinstance(arr, np.ndarray):
                return (arr.shape, arr.dtype, arr.flags["C_CONTIGUOUS"])
            return (type(arr), arr)

        for name, arr in args:
            logging.info(f"{name}: {_describe(arr)}")
    @profile
    def _predictor_corrector(
        self,
        pass_idx: int,
        image_a: np.ndarray,
        image_b: np.ndarray,
        interpolator="cubic",
        win_type="A",
    ):
        """Predictor-corrector step to adjust images based on previous displacement estimates."""

        n_win_y = len(self.win_ctrs_y[pass_idx])
        n_win_x = len(self.win_ctrs_x[pass_idx])
        self.delta_ab_pred = np.zeros((n_win_y, n_win_x, 2), dtype=np.float32)

        if pass_idx == 0:
            if self.delta_ab_old is None:
                self.delta_ab_old = np.zeros_like(self.delta_ab_pred)

            self.prev_win_size = (n_win_y, n_win_x)
            self.prev_win_spacing = (
                self.win_spacing_y[pass_idx],
                self.win_spacing_x[pass_idx],
            )
            return image_a.copy(), image_b.copy(), self.delta_ab_pred

        if self.delta_ab_old is None:
            raise RuntimeError("delta_ab_old is uninitialised before predictor step")

        interp_flag = cv2.INTER_CUBIC if interpolator == "cubic" else cv2.INTER_LINEAR

        self.delta_ab_old[..., 0] = gaussian_filter(
            self.delta_ab_old[..., 0],
            sigma=self.sd[pass_idx],
            truncate=(self.ksize_filt[pass_idx][0] - 1) / (2 * self.sd[pass_idx]),
            mode="nearest",
        )
        self.delta_ab_old[..., 1] = gaussian_filter(
            self.delta_ab_old[..., 1],
            sigma=self.sd[pass_idx],
            truncate=(self.ksize_filt[pass_idx][0] - 1) / (2 * self.sd[pass_idx]),
            mode="nearest",
        )

        self.delta_ab_dense = np.zeros((self.H, self.W, 2), dtype=np.float32)
        map_x_2d, map_y_2d = self.cached_dense_maps[pass_idx]
        if map_x_2d is None or map_y_2d is None:
            raise ValueError(f"Dense interpolation maps missing for pass {pass_idx}")
        
        # Verify cached dense maps have correct shape
        assert map_x_2d.shape == (self.H, self.W), f"Cached dense map X shape mismatch for pass {pass_idx}: {map_x_2d.shape} vs {(self.H, self.W)}"
        assert map_y_2d.shape == (self.H, self.W), f"Cached dense map Y shape mismatch for pass {pass_idx}: {map_y_2d.shape} vs {(self.H, self.W)}"
        logging.debug(f"Using cached dense interpolation maps for pass {pass_idx}")

        for d in range(2):
            self.delta_ab_dense[..., d] = cv2.remap(
                self.delta_ab_old[..., d].astype(np.float32),
                map_x_2d,
                map_y_2d,
                interp_flag,
                borderMode=cv2.BORDER_CONSTANT,
                borderValue=0,
            )

        delta_0b = self.delta_ab_dense / 2
        delta_0a = -delta_0b
        im_mesh_A = self.im_mesh + delta_0a
        im_mesh_B = self.im_mesh + delta_0b

        map_x, map_y = self.cached_predictor_maps[pass_idx]
        if map_x is None or map_y is None:
            raise ValueError(f"Predictor interpolation maps missing for pass {pass_idx}")
        
        # Verify cached predictor maps have correct shape
        expected_pred_shape = (len(self.win_ctrs_y[pass_idx]), len(self.win_ctrs_x[pass_idx]))
        assert map_x.shape == expected_pred_shape, f"Cached predictor map X shape mismatch for pass {pass_idx}: {map_x.shape} vs {expected_pred_shape}"
        assert map_y.shape == expected_pred_shape, f"Cached predictor map Y shape mismatch for pass {pass_idx}: {map_y.shape} vs {expected_pred_shape}"
        logging.debug(f"Using cached predictor interpolation maps for pass {pass_idx}")

        for d in range(2):
            remapped = cv2.remap(
                self.delta_ab_old[..., d].astype(np.float32),
                map_x,
                map_y,
                interp_flag,
                borderMode=cv2.BORDER_CONSTANT,
                borderValue=0.0,
            )
            self.delta_ab_pred[..., d] = remapped

        image_a_prime = cv2.remap(
            image_a.astype(np.float32),
            im_mesh_A[..., 1].astype(np.float32),
            im_mesh_A[..., 0].astype(np.float32),
            cv2.INTER_CUBIC,
            borderMode=cv2.BORDER_CONSTANT,
            borderValue=0,
        )
        image_b_prime = cv2.remap(
            image_b.astype(np.float32),
            im_mesh_B[..., 1].astype(np.float32),
            im_mesh_B[..., 0].astype(np.float32),
            cv2.INTER_CUBIC,
            borderMode=cv2.BORDER_CONSTANT,
            borderValue=0,
        )

        return image_a_prime, image_b_prime, self.delta_ab_pred
    @profile
    def _set_lib_arguments(
        self,
        config: Config,
        win_size: np.ndarray,
        pass_idx: int,
    ):
        """Set library arguments for PIV computation.

        :param config: Configuration object.
        :type config: Config
        :param win_size: Window size.
        :type win_size: np.ndarray
        :param pass_idx: Pass index.
        :type pass_idx: int
        :return: Tuple of library arguments.
        :rtype: tuple
        """
        # Window size: [win_height, win_width] in C-contiguous format
        win_size = np.ascontiguousarray(np.array(win_size, dtype=np.int32))

        n_win_y = len(self.win_ctrs_y[pass_idx])
        n_win_x = len(self.win_ctrs_x[pass_idx])
        # nWindows: [n_win_y, n_win_x] where n_win_y = rows, n_win_x = cols
        n_windows = np.ascontiguousarray(
            np.array([n_win_y, n_win_x], dtype=np.int32)
        )

        total_windows = n_win_y * n_win_x

        # Use precomputed vector mask for this pass if available
        # Mask shape: (n_win_y, n_win_x) in C-contiguous format
        if hasattr(self, 'vector_masks') and self.vector_masks and pass_idx < len(self.vector_masks):
            cached_mask = self.vector_masks[pass_idx]
            b_mask = np.ascontiguousarray(cached_mask.astype(np.float32))
        else:
            b_mask = np.ascontiguousarray(np.zeros((n_win_y, n_win_x), dtype=np.float32))
            logging.debug("No vector mask applied for pass %d", pass_idx)

        n_peaks = np.int32(config.num_peaks)
        i_peak_finder = np.int32(config.peak_finder)
        b_ensemble = bool(config.ensemble_piv)

        # Output arrays shape: (n_peaks, n_win_y, n_win_x) in C-contiguous format
        out_shape = (n_peaks, n_win_y, n_win_x)
        pk_loc_x = np.zeros(out_shape, dtype=np.float32)
        pk_loc_y = np.zeros(out_shape, dtype=np.float32)
        pk_height = np.zeros(out_shape, dtype=np.float32)
        sx = np.zeros(out_shape, dtype=np.float32)
        sy = np.zeros(out_shape, dtype=np.float32)
        sxy = np.zeros(out_shape, dtype=np.float32)

        # Correlation plane output: flattened array (not used, so use empty to save memory)
        correl_plane_out = np.empty(total_windows * win_size[0] * win_size[1], dtype=np.float32)

        if config.debug:
            args = [
                ("mask", b_mask),
                ("win_ctrs_x", self.win_ctrs_x[pass_idx].astype(np.float32)),
                ("win_ctrs_y", self.win_ctrs_y[pass_idx].astype(np.float32)),
                ("n_windows", n_windows),
                ("window_weight_a", self.win_weights[pass_idx]),
                ("b_ensemble", b_ensemble),
                ("window_weight_b", self.win_weights[pass_idx]),
                ("win_size", win_size),
                ("n_peaks", int(n_peaks)),
                ("i_peak_finder", int(i_peak_finder)),
                ("pk_loc_x", pk_loc_x),
                ("pk_loc_y", pk_loc_y),
                ("pk_height", pk_height),
                ("sx", sx),
                ("sy", sy),
                ("sxy", sxy),
                ("correl_plane_out", correl_plane_out),
            ]
            self._check_args(*args)

        return (
            win_size,
            n_windows,
            b_mask,
            n_peaks,
            i_peak_finder,
            b_ensemble,
            pk_loc_x,
            pk_loc_y,
            pk_height,
            sx,
            sy,
            sxy,
            correl_plane_out,
        )

    @profile
    def _cache_window_padding(self, config: Config) -> None:
        """Cache window padding information.

        :param config: Configuration object.
        :type config: Config
        """
        self.win_ctrs_x: list[np.ndarray] = []
        self.win_ctrs_y: list[np.ndarray] = []
        self.win_spacing_x: list[int] = []
        self.win_spacing_y: list[int] = []
        self.win_ctrs_x_all: list[np.ndarray] = []
        self.win_ctrs_y_all: list[np.ndarray] = []
        self.n_pre_all: list[tuple[int, int]] = []
        self.n_post_all: list[tuple[int, int]] = []
        self.ksize_filt: list[tuple[int, int]] = []
        self.sd: list[float] = []
        self.G_smooth_predictor: list[np.ndarray] = []

        H, W = config.image_shape

        for pass_idx, _ in enumerate(config.window_sizes):
            spacing_x, spacing_y, win_ctrs_x, win_ctrs_y = self._compute_window_centres(
                pass_idx, config
            )

            win_ctrs_x_pre = np.arange(1, win_ctrs_x[0] - spacing_x / 2, spacing_x)
            if win_ctrs_x_pre.size == 0:
                win_ctrs_x_pre = np.array([1])
            win_ctrs_x_pre -= 1
            win_ctrs_x_post = np.arange(
                W, win_ctrs_x[-1] + spacing_x / 2, -spacing_x
            )
            if win_ctrs_x_post.size == 0:
                win_ctrs_x_post = np.array([W])
            win_ctrs_x_post -= 1
            win_ctrs_x_all = np.concatenate(
                [win_ctrs_x_pre, win_ctrs_x, win_ctrs_x_post[::-1]]
            )

            win_ctrs_y_pre = np.arange(1, win_ctrs_y[0] - spacing_y / 2, spacing_y)
            if win_ctrs_y_pre.size == 0:
                win_ctrs_y_pre = np.array([1])
            win_ctrs_y_pre -= 1
            win_ctrs_y_post = np.arange(
                H, win_ctrs_y[-1] + spacing_y / 2, -spacing_y
            )
            if win_ctrs_y_post.size == 0:
                win_ctrs_y_post = np.array([H])
            win_ctrs_y_post -= 1
            win_ctrs_y_all = np.concatenate(
                [win_ctrs_y_pre, win_ctrs_y, win_ctrs_y_post[::-1]]
            )

            n_pre = (len(win_ctrs_y_pre), len(win_ctrs_x_pre))
            n_post = (len(win_ctrs_y_post), len(win_ctrs_x_post))

            self.win_ctrs_x.append(win_ctrs_x.astype(np.float32))
            self.win_ctrs_y.append(win_ctrs_y.astype(np.float32))
            self.win_spacing_x.append(spacing_x)
            self.win_spacing_y.append(spacing_y)
            self.win_ctrs_x_all.append(win_ctrs_x_all.astype(np.float32))
            self.win_ctrs_y_all.append(win_ctrs_y_all.astype(np.float32))
            self.n_pre_all.append(n_pre)
            self.n_post_all.append(n_post)

            if pass_idx == 0:
                self.ksize_filt.append((0, 0))
                self.sd.append(0.0)
                self.G_smooth_predictor.append(np.ones((1, 1), dtype=np.float32))
            else:
                prev_counts = (
                    len(self.win_ctrs_y[pass_idx - 1]),
                    len(self.win_ctrs_x[pass_idx - 1]),
                )
                prev_spacing = (
                    self.win_spacing_y[pass_idx - 1],
                    self.win_spacing_x[pass_idx - 1],
                )
                k_filt = (
                    np.round(np.array(prev_counts) / np.array(prev_spacing)).astype(int)
                    + 1
                )
                k_filt_list = [int(k) for k in k_filt.tolist()]
                k_filt_tuple = (
                    k_filt_list[0] + (k_filt_list[0] % 2 == 0),
                    k_filt_list[1] + (k_filt_list[1] % 2 == 0),
                )
                self.ksize_filt.append(k_filt_tuple)
                self.sd.append(np.sqrt(np.prod(k_filt_tuple)) / 3 * 0.65)
                g_kernel = self._window_weight_fun(k_filt_tuple, config.window_type)
                g_kernel = g_kernel.astype(np.float32)
                g_kernel /= max(np.sum(g_kernel), 1e-12)
                self.G_smooth_predictor.append(g_kernel)

        # # Verify window padding cache integrity
        # assert len(self.win_ctrs_x) == len(config.window_sizes), f"Window centers X cache length mismatch: {len(self.win_ctrs_x)} vs {len(config.window_sizes)}"
        # assert len(self.win_ctrs_y) == len(config.window_sizes), f"Window centers Y cache length mismatch: {len(self.win_ctrs_y)} vs {len(config.window_sizes)}"
        # assert len(self.win_spacing_x) == len(config.window_sizes), f"Window spacing X cache length mismatch: {len(self.win_spacing_x)} vs {len(config.window_sizes)}"
        # assert len(self.win_spacing_y) == len(config.window_sizes), f"Window spacing Y cache length mismatch: {len(self.win_spacing_y)} vs {len(config.window_sizes)}"
        
        # # Check that cached values are reasonable
        # for pass_idx in range(len(config.window_sizes)):
        #     assert len(self.win_ctrs_x[pass_idx]) > 0, f"No X window centers cached for pass {pass_idx}"
        #     assert len(self.win_ctrs_y[pass_idx]) > 0, f"No Y window centers cached for pass {pass_idx}"
        #     assert self.win_spacing_x[pass_idx] > 0, f"Invalid X spacing for pass {pass_idx}: {self.win_spacing_x[pass_idx]}"
        #     assert self.win_spacing_y[pass_idx] > 0, f"Invalid Y spacing for pass {pass_idx}: {self.win_spacing_y[pass_idx]}"
        
        logging.info(f"Successfully cached window padding for {len(config.window_sizes)} passes")

    @profile
    @profile
    def _cache_interpolation_grids(self, config: Config) -> None:
        """Cache interpolation grid coordinates for reuse across passes.

        This significantly improves performance by avoiding repeated
        computation of coordinate grids.
        
        :param config: Configuration object.
        :type config: Config
        """
        # Cache the image mesh for dense interpolation
        y_coords = np.arange(self.H, dtype=np.float32)
        x_coords = np.arange(self.W, dtype=np.float32)
        x_mesh, y_mesh = np.meshgrid(x_coords, y_coords)
        self.im_mesh = np.stack([y_mesh, x_mesh], axis=-1)
        
        # Pre-cache coordinate mappings for each pass
        self.cached_dense_maps = []
        self.cached_predictor_maps = []
        
        for pass_idx in range(len(config.window_sizes)):
            if pass_idx == 0:
                self.cached_dense_maps.append(None)
                self.cached_predictor_maps.append(None)
            else:
                # Cache dense interpolation maps
                points = (
                    self.win_ctrs_y_all[pass_idx - 1],
                    self.win_ctrs_x_all[pass_idx - 1]
                )
                map_x_1d = np.interp(
                    x_coords, points[1], np.arange(len(points[1]))
                )
                map_y_1d = np.interp(
                    y_coords, points[0], np.arange(len(points[0]))
                )
                map_x_2d, map_y_2d = np.meshgrid(
                    map_x_1d.astype(np.float32),
                    map_y_1d.astype(np.float32)
                )
                self.cached_dense_maps.append((map_x_2d, map_y_2d))
                
                # Cache predictor interpolation maps
                win_x, win_y = np.meshgrid(
                    self.win_ctrs_x[pass_idx],
                    self.win_ctrs_y[pass_idx]
                )
                ix = np.interp(
                    win_x.ravel(), points[1], np.arange(len(points[1]))
                )
                iy = np.interp(
                    win_y.ravel(), points[0], np.arange(len(points[0]))
                )
                map_x = ix.reshape(win_x.shape).astype(np.float32)
                map_y = iy.reshape(win_x.shape).astype(np.float32)
                self.cached_predictor_maps.append((map_x, map_y))

        # Verify caching integrity
        assert len(self.cached_dense_maps) == len(config.window_sizes), f"Dense maps cache length mismatch: {len(self.cached_dense_maps)} vs {len(config.window_sizes)}"
        assert len(self.cached_predictor_maps) == len(config.window_sizes), f"Predictor maps cache length mismatch: {len(self.cached_predictor_maps)} vs {len(config.window_sizes)}"
        
        # Check that non-zero passes have cached maps
        for pass_idx in range(1, len(config.window_sizes)):
            assert self.cached_dense_maps[pass_idx] is not None, f"Dense map for pass {pass_idx} is None"
            assert self.cached_predictor_maps[pass_idx] is not None, f"Predictor map for pass {pass_idx} is None"
            dense_x, dense_y = self.cached_dense_maps[pass_idx]
            pred_x, pred_y = self.cached_predictor_maps[pass_idx]
            assert dense_x.shape == (self.H, self.W), f"Dense map X shape incorrect for pass {pass_idx}: {dense_x.shape} vs {(self.H, self.W)}"
            assert dense_y.shape == (self.H, self.W), f"Dense map Y shape incorrect for pass {pass_idx}: {dense_y.shape} vs {(self.H, self.W)}"
            expected_pred_shape = (len(self.win_ctrs_y[pass_idx]), len(self.win_ctrs_x[pass_idx]))
            assert pred_x.shape == expected_pred_shape, f"Predictor map X shape incorrect for pass {pass_idx}: {pred_x.shape} vs {expected_pred_shape}"
            assert pred_y.shape == expected_pred_shape, f"Predictor map Y shape incorrect for pass {pass_idx}: {pred_y.shape} vs {expected_pred_shape}"
        
        logging.info(f"Successfully cached interpolation grids for {len(config.window_sizes)} passes")

    # def _apply_mask_to_vectors(
    #     self,
    #     win_ctrs_x: np.ndarray,
    #     win_ctrs_y: np.ndarray,
    #     mask: np.ndarray
    # ) -> np.ndarray:
    #     """
    #     Apply user-defined mask to invalidate vectors in masked regions.
        
    #     A vector is invalidated if its window center falls within a masked region
    #     (where mask == True).
        
    #     Parameters
    #     ----------
    #     win_ctrs_x : np.ndarray
    #         1D array of window center x-coordinates
    #     win_ctrs_y : np.ndarray
    #         1D array of window center y-coordinates
    #     mask : np.ndarray
    #         Boolean mask array of shape (H, W) where True indicates masked regions
            
    #     Returns
    #     -------
    #     np.ndarray
    #         Boolean mask of shape (len(win_ctrs_y), len(win_ctrs_x)) where
    #         True indicates vectors to invalidate
    #     """
    #     grid_y, grid_x = np.meshgrid(win_ctrs_y, win_ctrs_x, indexing="ij")

    #     # Round to nearest pixel indices
    #     x_idx = np.round(grid_x).astype(int)
    #     y_idx = np.round(grid_y).astype(int)

    #     # Clip to valid image bounds
    #     x_idx = np.clip(x_idx, 0, mask.shape[1] - 1)
    #     y_idx = np.clip(y_idx, 0, mask.shape[0] - 1)

    #     # Sample mask at window center locations
    #     # mask[y, x] where True = masked region
    #     vector_mask = mask[y_idx, x_idx]

    #     return vector_mask



import threading
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Optional

import numpy as np
from flask import Blueprint, jsonify, request, send_file
from loguru import logger
from scipy.io import loadmat

from pivtools_core.config import get_config
from pivtools_core.paths import get_data_paths
from ..pod_decompose import pod_decompose

POD_bp = Blueprint("POD", __name__)


# In-memory POD job state
_pod_state: Dict[str, Any] = {
    "processing": False,
    "progress": 0,  # 0..100 (coarse: 0 at start, 100 at end)
    "message": None,
    "started_at": None,
    "finished_at": None,
    "error": None,
}
_pod_thread: Optional[threading.Thread] = None
_pod_cancel_event = threading.Event()
_state_lock = threading.Lock()


def _set_state(**kwargs):
    with _state_lock:
        _pod_state.update(kwargs)


def _reset_state():
    _set_state(
        processing=False,
        progress=0,
        message=None,
        started_at=None,
        finished_at=None,
        error=None,
    )


def _run_pod_job(base: Path, cam: int):
    """Background job entry to run POD and update state."""
    cfg = get_config(refresh=True)
    try:
        _set_state(
            processing=True,
            progress=0,
            started_at=datetime.utcnow().isoformat(),
            message="POD running",
            error=None,
        )
        logger.info(f"[POD] Starting POD job | base='{base}', cam={cam}")
        # Determine k_modes from config (if present) else default 10
        k_modes = 10
        try:
            for entry in cfg.post_processing or []:
                if entry.get("type") == "POD":
                    k_modes = int(entry.get("settings", {}).get("k_modes", k_modes))
                    break
        except Exception:
            pass

        # Note: Cancellation is best-effort; pod_decompose does not currently poll cancel event.
        pod_decompose(cam_num=int(cam), config=cfg, base=Path(base), k_modes=k_modes)
        _set_state(
            progress=100,
            message="POD completed",
            processing=False,
            finished_at=datetime.utcnow().isoformat(),
        )
        logger.info("[POD] Job completed successfully")
    except Exception as e:
        logger.exception(f"[POD] Job failed: {e}")
        _set_state(
            processing=False,
            error=str(e),
            message="POD failed",
            finished_at=datetime.utcnow().isoformat(),
        )


@POD_bp.route("/start_pod", methods=["POST"])
def start_pod():
    """Start POD in a background thread using current config and provided selection.

    Expects JSON payload with optional fields:
      - basepath_idx: int (index into config.base_paths)
      - base_path: str (absolute path to base directory; takes precedence if provided)
      - camera: int or str (camera number)
    """
    global _pod_thread

    # If a job is already running, do not start another
    if _pod_thread is not None and _pod_thread.is_alive():
        with _state_lock:
            st = {k: _pod_state.get(k) for k in ("processing", "progress", "message")}
        return (
            jsonify({"status": "busy", **st}),
            409,
        )

    data = request.get_json(silent=True) or {}
    cfg = get_config(refresh=True)

    # Resolve base directory
    base_path_str = data.get("base_path")
    base: Path
    if isinstance(base_path_str, str) and base_path_str.strip():
        base = Path(base_path_str).expanduser()
    else:
        idx = int(data.get("basepath_idx", 0))
        try:
            base = cfg.base_paths[idx]
        except Exception:
            base = cfg.base_paths[0]

    # Resolve camera
    cam_raw = data.get("camera")
    try:
        cam = int(cam_raw) if cam_raw is not None else int(cfg.camera_numbers[0])
    except Exception:
        cam = int(cfg.camera_numbers[0])

    # Reset and start job
    _pod_cancel_event.clear()
    _reset_state()
    _set_state(message="POD queued")

    _pod_thread = threading.Thread(target=_run_pod_job, args=(base, cam), daemon=True)
    _pod_thread.start()

    return jsonify({"status": "started", "processing": True, "progress": 0}), 202


@POD_bp.route("/cancel_pod", methods=["POST"])
def cancel_pod():
    """Signal cancellation of a running POD job.

    Note: Current compute path does not cooperatively check for cancellation, so this is best-effort.
    We still mark state as cancelling; the job thread may continue until current computation completes.
    """
    _pod_cancel_event.set()
    with _state_lock:
        is_running = bool(_pod_thread is not None and _pod_thread.is_alive())
    if is_running:
        _set_state(message="Cancellation requested")
        return jsonify({"status": "cancelling", "processing": True}), 202
    _reset_state()
    return jsonify({"status": "idle", "processing": False}), 200


@POD_bp.route("/pod_status", methods=["GET"])
def pod_status():
    """Return current POD job status suitable for frontend polling."""
    with _state_lock:
        st = dict(_pod_state)
        st["processing"] = bool(
            st.get("processing", False)
            or (_pod_thread is not None and _pod_thread.is_alive())
        )
    # Keep progress within [0,100]
    try:
        st["progress"] = int(max(0, min(100, int(st.get("progress", 0)))))
    except Exception:
        st["progress"] = 0
    # Provide numeric 'status' for legacy clients expecting it
    st["status"] = st["progress"]
    return jsonify(st), 200


@POD_bp.route("/pod_energy", methods=["GET"])
def pod_energy():
    """Return modal energy breakdown from saved POD results for a given run.

    Query params:
      - base_path or basepath_idx
      - camera (int)
      - run (int, 1-based; default 1)
      - merged ("1"/"0")
    """
    cfg = get_config(refresh=True)

    # Resolve base directory
    base_path_str = request.args.get("base_path")
    if base_path_str and base_path_str.strip():
        base = Path(base_path_str).expanduser()
    else:
        try:
            idx = int(request.args.get("basepath_idx", 0))
        except Exception:
            idx = 0
        try:
            base = cfg.base_paths[idx]
        except Exception:
            base = cfg.base_paths[0]

    # Resolve camera
    try:
        cam = int(request.args.get("camera", cfg.camera_numbers[0]))
    except Exception:
        cam = int(cfg.camera_numbers[0])

    # Run label
    try:
        run_label = int(request.args.get("run", 1))
    except Exception:
        run_label = 1

    # Merged flag
    merged_flag = request.args.get("merged", "0") in ("1", "true", "True")

    # Find POD settings for endpoint/source_type
    endpoint = ""
    source_type = "instantaneous"
    try:
        for entry in cfg.post_processing or []:
            if entry.get("type") == "POD":
                s = entry.get("settings", {}) or {}
                endpoint = entry.get("endpoint", s.get("endpoint", "")) or ""
                source_type = (
                    entry.get("source_type", s.get("source_type", "instantaneous"))
                    or "instantaneous"
                )
                break
    except Exception:
        pass

    # Stats directory (same logic as in pod_decompose)
    paths = get_data_paths(
        base_dir=base,
        num_images=cfg.num_images,
        cam=cam,
        type_name=source_type,
        endpoint=endpoint,
        use_merged=merged_flag,
    )
    stats_base = paths["stats_dir"]

    run_dir_rand = stats_base / "pod_randomised" / f"run_{run_label:02d}"
    run_dir_exact = stats_base / "POD" / f"run_{run_label:02d}"

    joint_file = "POD_joint.mat"
    sep_file = "POD_separate.mat"

    stats_path = None
    stacked = None
    for base_dir in (run_dir_rand, run_dir_exact):
        if (base_dir / joint_file).exists():
            stats_path = base_dir / joint_file
            stacked = True
            break
        if (base_dir / sep_file).exists():
            stats_path = base_dir / sep_file
            stacked = False
            break

    if stats_path is None:
        return jsonify({"error": f"No POD stats found for run {run_label}"}), 404

    try:
        mat = loadmat(str(stats_path), struct_as_record=False, squeeze_me=True)
        # Minimal meta extraction for JSON serialisation

        def _get_meta(meta_obj, key, default=None):
            try:
                if isinstance(meta_obj, dict):
                    return meta_obj.get(key, default)
                return getattr(meta_obj, key, default)
            except Exception:
                return default

        meta_obj = mat.get("meta", {})
        meta = {
            "run_label": int(_get_meta(meta_obj, "run_label", run_label)),
            "cam": int(_get_meta(meta_obj, "cam", cam)),
            "endpoint": _get_meta(meta_obj, "endpoint", endpoint),
            "source_type": _get_meta(meta_obj, "source_type", source_type),
            "stack_U_y": bool(_get_meta(meta_obj, "stack_U_y", bool(stacked))),
            "normalise": bool(_get_meta(meta_obj, "normalise", False)),
            "algorithm": _get_meta(meta_obj, "algorithm", "exact"),
        }

        if stacked:
            ef = np.asarray(mat.get("energy_fraction", []), dtype=float).ravel()
            ec = np.asarray(mat.get("energy_cumulative", []), dtype=float).ravel()
            k = int(ef.size)
            return (
                jsonify(
                    {
                        "stacked": True,
                        "k": k,
                        "energy_fraction": ef.tolist(),
                        "energy_cumulative": ec.tolist(),
                        "meta": meta,
                    }
                ),
                200,
            )
        else:
            ef_u = np.asarray(mat.get("energy_fraction_ux", []), dtype=float).ravel()
            ec_u = np.asarray(mat.get("energy_cumulative_ux", []), dtype=float).ravel()
            ef_v = np.asarray(mat.get("energy_fraction_uy", []), dtype=float).ravel()
            ec_v = np.asarray(mat.get("energy_cumulative_uy", []), dtype=float).ravel()
            k = int(max(ef_u.size, ef_v.size))
            return (
                jsonify(
                    {
                        "stacked": False,
                        "k": k,
                        "energy_fraction_ux": ef_u.tolist(),
                        "energy_cumulative_ux": ec_u.tolist(),
                        "energy_fraction_uy": ef_v.tolist(),
                        "energy_cumulative_uy": ec_v.tolist(),
                        "meta": meta,
                    }
                ),
                200,
            )
    except Exception as e:
        logger.exception(f"[POD] Failed to read energy from {stats_path}: {e}")
        return jsonify({"error": str(e)}), 500


@POD_bp.route("/pod_energy_modes", methods=["GET"])
def pod_energy_modes():
    """
    Return modal energy breakdown (fraction and cumulative) for all modes for a given run.
    Query params:
      - base_path or basepath_idx
      - camera (int)
      - run (int, 1-based; default 1)
      - merged ("1"/"0")
    """
    cfg = get_config(refresh=True)

    # Resolve base directory
    base_path_str = request.args.get("base_path")
    if base_path_str and base_path_str.strip():
        base = Path(base_path_str).expanduser()
    else:
        try:
            idx = int(request.args.get("basepath_idx", 0))
        except Exception:
            idx = 0
        try:
            base = cfg.base_paths[idx]
        except Exception:
            base = cfg.base_paths[0]

    # Resolve camera
    try:
        cam = int(request.args.get("camera", cfg.camera_numbers[0]))
    except Exception:
        cam = int(cfg.camera_numbers[0])

    # Run label
    try:
        run_label = int(request.args.get("run", 1))
    except Exception:
        run_label = 1

    # Merged flag
    merged_flag = request.args.get("merged", "0") in ("1", "true", "True")

    # Find POD settings for endpoint/source_type
    endpoint = ""
    source_type = "instantaneous"
    try:
        for entry in cfg.post_processing or []:
            if entry.get("type") == "POD":
                s = entry.get("settings", {}) or {}
                endpoint = entry.get("endpoint", s.get("endpoint", "")) or ""
                source_type = (
                    entry.get("source_type", s.get("source_type", "instantaneous"))
                    or "instantaneous"
                )
                break
    except Exception:
        pass

    # Stats directory (same logic as in pod_decompose)
    paths = get_data_paths(
        base_dir=base,
        num_images=cfg.num_images,
        cam=cam,
        type_name=source_type,
        endpoint=endpoint,
        use_merged=merged_flag,
    )
    stats_base = paths["stats_dir"]

    # --- Search for POD energy summary in both randomised and exact directories ---
    run_dir_rand = stats_base / "pod_randomised" / f"run_{run_label:02d}"
    run_dir_exact = stats_base / "POD" / f"run_{run_label:02d}"

    # Check for the energy summary file in both locations
    summary_path = None
    stacked = None
    # Try randomised first, then exact
    for base_dir in (run_dir_rand, run_dir_exact):
        f = base_dir / "POD_energy_summary.mat"
        if f.exists():
            summary_path = f
            # Heuristic: if "energy_fraction" present, it's stacked; else separate
            try:
                mat = loadmat(str(f), struct_as_record=False, squeeze_me=True)
                stacked = "energy_fraction" in mat
                break
            except Exception:
                continue

    # If summary not found, try the joint/separate files
    if summary_path is None:
        for base_dir in (run_dir_rand, run_dir_exact):
            for file_name in ["POD_joint.mat", "POD_separate.mat"]:
                f = base_dir / file_name
                if f.exists():
                    summary_path = f
                    stacked = file_name == "POD_joint.mat"
                    break
            if summary_path is not None:
                break

    if summary_path is None:
        # Diagnostic: show where we looked
        logger.error(
            f"[POD] Could not find energy data in: {run_dir_rand} or {run_dir_exact}"
        )
        return jsonify({"error": f"No POD energy data found for run {run_label}"}), 404

    try:
        # If client only wants headers (HEAD) return quickly to avoid expensive loadmat/json work
        if request.method == "HEAD":
            return ("", 200)

        mat = loadmat(str(summary_path), struct_as_record=False, squeeze_me=True)
        meta_obj = mat.get("meta", {})

        # Minimal meta extraction for JSON serialisation
        def _get_meta(meta_obj, key, default=None):
            try:
                if isinstance(meta_obj, dict):
                    return meta_obj.get(key, default)
                return getattr(meta_obj, key, default)
            except Exception:
                return default

        # Normalize meta fields to native Python types (avoid numpy ndarrays etc.)
        meta = {
            "run_label": int(_get_meta(meta_obj, "run_label", run_label) or run_label),
            "cam": int(_get_meta(meta_obj, "cam", cam) or cam),
            "endpoint": str(_get_meta(meta_obj, "endpoint", endpoint) or endpoint),
            "source_type": str(
                _get_meta(meta_obj, "source_type", source_type) or source_type
            ),
            "stack_U_y": bool(_get_meta(meta_obj, "stack_U_y", bool(stacked))),
            "normalise": bool(_get_meta(meta_obj, "normalise", False)),
            "algorithm": str(_get_meta(meta_obj, "algorithm", "exact") or "exact"),
        }

        # Add more diagnostics to the meta information
        meta["file_path"] = str(summary_path)
        meta["file_name"] = summary_path.name

        if stacked:
            ef = np.asarray(mat.get("energy_fraction", []), dtype=float).ravel()
            ec = np.asarray(mat.get("energy_cumulative", []), dtype=float).ravel()

            # Add useful summary statistics
            total_modes = len(ef)
            threshold_95 = (
                next((i + 1 for i, v in enumerate(ec) if v >= 0.95), total_modes)
                if len(ec) > 0
                else None
            )
            threshold_99 = (
                next((i + 1 for i, v in enumerate(ec) if v >= 0.99), total_modes)
                if len(ec) > 0
                else None
            )

            return (
                jsonify(
                    {
                        "stacked": True,
                        "energy_fraction": ef.tolist(),
                        "energy_cumulative": ec.tolist(),
                        "meta": meta,
                        "summary": {
                            "total_modes": total_modes,
                            "modes_for_95_percent": threshold_95,
                            "modes_for_99_percent": threshold_99,
                            "first_mode_energy": float(ef[0]) if len(ef) > 0 else 0,
                        },
                    }
                ),
                200,
            )
        else:
            ef_u = np.asarray(mat.get("energy_fraction_ux", []), dtype=float).ravel()
            ec_u = np.asarray(mat.get("energy_cumulative_ux", []), dtype=float).ravel()
            ef_v = np.asarray(mat.get("energy_fraction_uy", []), dtype=float).ravel()
            ec_v = np.asarray(mat.get("energy_cumulative_uy", []), dtype=float).ravel()

            # Add useful summary statistics
            total_modes_u = len(ef_u)
            total_modes_v = len(ef_v)
            threshold_95_u = (
                next((i + 1 for i, v in enumerate(ec_u) if v >= 0.95), total_modes_u)
                if len(ec_u) > 0
                else None
            )
            threshold_95_v = (
                next((i + 1 for i, v in enumerate(ec_v) if v >= 0.95), total_modes_v)
                if len(ec_v) > 0
                else None
            )
            threshold_99_u = (
                next((i + 1 for i, v in enumerate(ec_u) if v >= 0.99), total_modes_u)
                if len(ec_u) > 0
                else None
            )
            threshold_99_v = (
                next((i + 1 for i, v in enumerate(ec_v) if v >= 0.99), total_modes_v)
                if len(ec_v) > 0
                else None
            )

            return (
                jsonify(
                    {
                        "stacked": False,
                        "energy_fraction_ux": ef_u.tolist(),
                        "energy_cumulative_ux": ec_u.tolist(),
                        "energy_fraction_uy": ef_v.tolist(),
                        "energy_cumulative_uy": ec_v.tolist(),
                        "meta": meta,
                        "summary_ux": {
                            "total_modes": total_modes_u,
                            "modes_for_95_percent": threshold_95_u,
                            "modes_for_99_percent": threshold_99_u,
                            "first_mode_energy": float(ef_u[0]) if len(ef_u) > 0 else 0,
                        },
                        "summary_uy": {
                            "total_modes": total_modes_v,
                            "modes_for_95_percent": threshold_95_v,
                            "modes_for_99_percent": threshold_99_v,
                            "first_mode_energy": float(ef_v[0]) if len(ef_v) > 0 else 0,
                        },
                    }
                ),
                200,
            )
    except Exception as e:
        logger.exception(f"[POD] Failed to read energy data from {summary_path}: {e}")
        return jsonify({"error": str(e)}), 500


@POD_bp.route("/pod_energy_png", methods=["GET"])
def pod_energy_png():
    """Return the cumulative POD energy PNG file for a given run if present.

    Query params: base_path or basepath_idx, camera, run, merged
    """
    cfg = get_config(refresh=True)

    # Resolve base directory
    base_path_str = request.args.get("base_path")
    if base_path_str and base_path_str.strip():
        base = Path(base_path_str).expanduser()
    else:
        try:
            idx = int(request.args.get("basepath_idx", 0))
        except Exception:
            idx = 0
        try:
            base = cfg.base_paths[idx]
        except Exception:
            base = cfg.base_paths[0]

    # Resolve camera
    try:
        cam = int(request.args.get("camera", cfg.camera_numbers[0]))
    except Exception:
        cam = int(cfg.camera_numbers[0])

    # Run label
    try:
        run_label = int(request.args.get("run", 1))
    except Exception:
        run_label = 1

    merged_flag = request.args.get("merged", "0") in ("1", "true", "True")

    # Find POD endpoint settings
    endpoint = ""
    source_type = "instantaneous"
    try:
        for entry in cfg.post_processing or []:
            if entry.get("type") == "POD":
                s = entry.get("settings", {}) or {}
                endpoint = entry.get("endpoint", s.get("endpoint", "")) or ""
                source_type = (
                    entry.get("source_type", s.get("source_type", "instantaneous"))
                    or "instantaneous"
                )
                break
    except Exception:
        pass

    paths = get_data_paths(
        base_dir=base,
        num_images=cfg.num_images,
        cam=cam,
        type_name=source_type,
        endpoint=endpoint,
        use_merged=merged_flag,
    )
    stats_base = paths["stats_dir"]

    run_dir_rand = stats_base / "pod_randomised" / f"run_{run_label:02d}"
    run_dir_exact = stats_base / "POD" / f"run_{run_label:02d}"

    candidates = []
    # try with config.plot_save_extension if available, else .png
    ext = getattr(cfg, "plot_save_extension", ".png") or ".png"
    candidates.append(run_dir_rand / f"POD_energy_cumulative{ext}")
    candidates.append(run_dir_exact / f"POD_energy_cumulative{ext}")
    candidates.append(run_dir_rand / "POD_energy_cumulative.png")
    candidates.append(run_dir_exact / "POD_energy_cumulative.png")

    for f in candidates:
        if f.exists():
            try:
                return send_file(str(f), mimetype="image/png")
            except Exception as e:
                logger.exception(f"[POD] Failed to send PNG {f}: {e}")
                break

    return jsonify({"error": "POD cumulative PNG not found"}), 404
